@@ -6,7 +6,7 @@
 #                            | (__| |_| |  _ <| |___
 #                             \___|\___/|_| \_\_____|
 #
-# Copyright (C) 1998 - 2022, Daniel Stenberg, <daniel@haxx.se>, et al.
+# Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution. The terms
@@ -19,6 +19,8 @@
 # This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 # KIND, either express or implied.
 #
+# SPDX-License-Identifier: curl
+#
 ###########################################################################
 #
 #       curl compilation script for the OS/400.
@@ -35,7 +37,7 @@ cd "${TOPDIR}"
 
 if action_needed "${LIBIFSNAME}"
 then    CMD="CRTLIB LIB(${TARGETLIB}) TEXT('curl: multiprotocol support API')"
-        system "${CMD}"
+        CLcommand "${CMD}"
 fi
 
 
@@ -44,7 +46,7 @@ fi
 if action_needed "${LIBIFSNAME}/DOCS.FILE"
 then    CMD="CRTSRCPF FILE(${TARGETLIB}/DOCS) RCDLEN(240)"
         CMD="${CMD} CCSID(${TGTCCSID}) TEXT('Documentation texts')"
-        system "${CMD}"
+        CLcommand "${CMD}"
 fi
 
 
@@ -64,11 +66,55 @@ do      MEMBER="`basename \"${TEXT}\" .OS400`"
         if action_needed "${MEMBER}" "${TEXT}"
         then    CMD="CPY OBJ('${TEXT}') TOOBJ('${MEMBER}') TOCCSID(${TGTCCSID})"
                 CMD="${CMD} DTAFMT(*TEXT) REPLACE(*YES)"
-                system "${CMD}"
+                CLcommand "${CMD}"
         fi
 done
 
 
+#       Create the RPGXAMPLES source file if it does not exist.
+
+if action_needed "${LIBIFSNAME}/RPGXAMPLES.FILE"
+then    CMD="CRTSRCPF FILE(${TARGETLIB}/RPGXAMPLES) RCDLEN(240)"
+        CMD="${CMD} CCSID(${TGTCCSID}) TEXT('ILE/RPG examples')"
+        CLcommand "${CMD}"
+fi
+
+
+#       Copy RPG examples if needed.
+
+for EXAMPLE in "${SCRIPTDIR}/rpg-examples"/*
+do      MEMBER="`basename \"${EXAMPLE}\"`"
+        IFSMEMBER="${LIBIFSNAME}/RPGXAMPLES.FILE/`db2_name \"${MEMBER}\"`.MBR"
+
+        [ -e "${EXAMPLE}" ] || continue
+
+        if action_needed "${IFSMEMBER}" "${EXAMPLE}"
+        then    CMD="CPY OBJ('${EXAMPLE}') TOOBJ('${IFSMEMBER}')"
+                CMD="${CMD} TOCCSID(${TGTCCSID}) DTAFMT(*TEXT) REPLACE(*YES)"
+                CLcommand "${CMD}"
+                MBRTEXT=`sed -e '1!d;/^      \*/!d;s/^ *\* *//'         \
+                             -e 's/ *$//;s/'"'"'/&&/g' < "${EXAMPLE}"`
+                CMD="CHGPFM FILE(${TARGETLIB}/RPGXAMPLES) MBR(${MEMBER})"
+                CMD="${CMD} SRCTYPE(RPGLE) TEXT('${MBRTEXT}')"
+                CLcommand "${CMD}"
+        fi
+done
+
+
+#       Compile the QADRTMAIN2 replacement module.
+
+if action_needed "${LIBIFSNAME}/CURLMAIN.MODULE" "${SCRIPTDIR}/curlmain.c"
+then    CMD="CRTCMOD MODULE(${TARGETLIB}/CURLMAIN)"
+        CMD="${CMD} SRCSTMF('${SCRIPTDIR}/curlmain.c')"
+        CMD="${CMD} SYSIFCOPT(*IFS64IO) LOCALETYPE(*LOCALE) FLAG(10)"
+        CMD="${CMD} TGTCCSID(${TGTCCSID}) TGTRLS(${TGTRLS})"
+        CMD="${CMD} OUTPUT(${OUTPUT})"
+        CMD="${CMD} OPTIMIZE(${OPTIMIZE})"
+        CMD="${CMD} DBGVIEW(${DEBUG})"
+        CLcommand "${CMD}"
+fi
+
+
 #       Build in each directory.
 
 # for SUBDIR in include lib src tests
